from abc import abstractmethod
from contextlib import contextmanager
from typing import ContextManager, Callable, Union, Optional, Dict, Any, List


TError = Union[
    str,
    List[str],
    Dict[str, Any],
    'ValidationError',
]


class ValidationError(Exception):
    """
    TODO
    """

    def __init__(self, error: Optional[TError] = None):
        """
        :param error:
        """
        if isinstance(error, ValidationError):
            error = error.error
        self.error = error or {}

    def __str__(self):
        return str(self.error)

    # @abstractmethod
    # def __repr__(self):
    #     """
    #     :return:
    #     """
    #     return 'ValidationError'

    def __bool__(self) -> bool:
        """
        :return:
        """
        return bool(self.error)

    @abstractmethod
    def add(
            self,
            key: str,
            error: TError,
    ):
        """
        :param key:
        :param error:
        """
        # if isinstance()

        if isinstance(error, str):
            self.error.setdefault(key, []).append(error)
        elif isinstance(error, ValidationError):
            self.error[key] = error.error
        else:
            raise ValueError(f'Invalid type: {type(error)}')


# class ValidationError(Exception):
#     """
#     TODO
#     """
#
#     def __init__(self, errors: Optional[Dict[str, List[str]]] = None):
#         """
#         :param errors:
#         """
#         self.errors = errors or {}
#
#     def __str__(self):
#         return str(self.errors)
#
#     def __bool__(self) -> bool:
#         """
#         :return:
#         """
#         return bool(self.errors)
#
#     def add(
#             self,
#             key: str,
#             error: Union[str, 'ValidationError'],
#     ):
#         """
#         :param key:
#         :param error:
#         """
#         if isinstance(error, str):
#             self.errors.setdefault(key, []).append(error)
#         elif isinstance(error, ValidationError):
#             self.errors[key] = error
#         else:
#             raise ValueError(f'Invalid type: {type(error)}')


TValidator = Callable


class Validator(object):
    pass


# class ValidationErrorGroup(object):
#
#     def __init__(self):
#         self.errors = ValidationError()
#
#     def __enter__(self) -> ValidationError:
#         return self.errors
#
#     def __exit__(self, exc_type, exc_value, exc_traceback):
#         if self.errors:
#             raise self.errors


@contextmanager
def collect_validation_errors() -> ContextManager[ValidationError]:
    """
    A decorator to accumulate errors.
    """
    errors = ValidationError()
    yield errors
    if errors:
        raise errors



from typing import List, Optional, Type, TypeVar, Generic, Any, Union, \
    Callable

from .context import Context, default_context
from .validation import TValidator, ValidationError


# -- Types -------------------------------------------------------------------


# Type of deserialized (Python) objects
TObject = TypeVar('TObject')

# Type of serialized data
TSerialized = TypeVar('TSerialized')


# -- Field -------------------------------------------------------------------


class Field(Generic[TObject, TSerialized]):
    """
    TODO
    """

    # Shortcut
    ValidationError = ValidationError

    def __init__(
            self,
            cls: Type[TObject],
            required: bool = True,
            nullable: bool = False,
            cast: bool = True,
            key: Optional[str] = None,
            attr: Optional[str] = None,
            default: Optional[Union[TObject, Callable[[], TObject]]] = None,
            validators: Optional[List[TValidator]] = None,
            serializable: bool = True,
            deserializable: bool = True,
            name: Optional[str] = None,
            description: Optional[str] = None,
    ):
        """
        :param cls: The class to instantiate when deserializing data into
            objects. Will receive.
        :param required: Whether this field is required (can not be None)
        :param nullable:
        :param cast: Allow casting of value when deserializing.
            Otherwise, not-exact datatypes causes ValidationError.
        :param key:
        :param attr:
        :param default: Default value, or a callable.
        :param validators:
        :param serializable:
        :param deserializable:
        :param name:
        :param description:
        """
        self.cls = cls
        self.required = required
        # self.nullable = nullable
        # self.cast = cast
        self.key = key
        self.attr = attr
        self.default = default
        self.validators = validators or []
        self.serializable = serializable
        self.deserializable = deserializable
        self.name = name
        self.description = description

    @property
    def default_value(self) -> Any:
        """
        :returns: Field default value, including None
        """
        if callable(self.default):
            return self.default()
        else:
            return self.default

    def cast(self, value: Any, context: Context) -> TObject:
        return self.cls(value)
        # try:
        #     data = self.cast(data, context)
        # except (ValueError, TypeError):
        #     raise ValidationError(f'Can not cast to {self.cls}: {data}')

    def normalize(self, value: Any, context: Context) -> TObject:
        return value

    # ------------------------------------------------------------------------

    def serialize(
            self,
            obj: TObject,
            context: Context = default_context,
    ) -> TSerialized:
        """
        :param obj:
        :param context:
        :return:
        """
        if self.serializable is not True:
            raise RuntimeError('Serializing un-serializable field')

        if self.required:
            if obj is None:
                raise ValidationError('Field can not be null')

        if obj is not None and not isinstance(obj, self.cls):
            try:
                obj = self.cast(obj, context)
            except (ValueError, TypeError):
                raise ValidationError(f'Can not cast to {self.cls}: {obj}')

        return obj

    def deserialize(
            self,
            data: TSerialized,
            context: Context = default_context,
    ) -> TObject:
        """
        :param data:
        :param context:
        :return:
        """
        if self.serializable is not True:
            raise RuntimeError('Serializing un-serializable field')

        if self.required:
            if data is None:
                raise ValidationError('Field can not be null')

        if data is not None and not isinstance(data, self.cls):
            # data = self.cast(data, context)
            try:
                data = self.cast(data, context)
            except (ValueError, TypeError):
                raise ValidationError(f'Can not cast to {self.cls}: {data}')

        return data

        # if self.deserializable is not True:
        #     raise RuntimeError('Deserializing un-deserializable field')
        #
        # return self.deserializer(
        #     value=data,
        #     context=context,
        # )


# class SimpleField(Field[TObject, TSerialized]):
#
#     def __init__(self, cls: Type[TObject], **kwargs):
#         """
#         :param cls: The class to instantiate when deserializing data into
#             objects. Will receive.
#         :param kwargs: **kwargs for Field.__init__()
#         """
#         super().__init__(**kwargs)
#         self.cls = cls
#
#     # @cached_property
#     @property
#     def serializer(self) -> Pipeline:
#         """
#         :returns:
#         """
#         l = 2
#         pipe = [Cast(self.cls)]
#         pipe.extend(self.validators)
#
#         j = 2
#
#         return Pipeline(*pipe)
#
#     # @cached_property
#     @property
#     def deserializer(self) -> Pipeline:
#         """
#         :returns:
#         """
#         pipe = [Cast(self.cls)]
#         pipe.extend(self.validators)
#
#         j = 2
#
#         return Pipeline(*pipe)


# -- Transformations ---------------------------------------------------------


class SetDefault(object):
    def __init__(self, default: Any):
        self.default = default

    def __call__(self, value: Any, context: Context) -> Any:
        if value is None:
            value = self.default
        return value


# class


# class FieldOLD(Generic[TObject, TSerialized]):
#     """
#     TODO
#     """
#
#     @classmethod
#     def create(cls, context: Context, data: TSerialized) -> TObject:
#         """
#         :param context:
#         :param data:
#         :returns:
#         """
#         return cls(data)
#
#     def __init__(
#             self,
#             cls: Optional[Type[TObject]] = None,
#             required: bool = True,
#             nullable: bool = False,
#             cast: bool = True,
#             key: Optional[str] = None,
#             attr: Optional[str] = None,
#             default: Optional[TObject] = None,
#             validators: Optional[List[TValidator]] = None,
#             serializable: bool = True,
#             deserializable: bool = True,
#             description: Optional[str] = None,
#     ):
#         """
#         :param cls: The class to instantiate when deserializing data into
#             objects. Will receive.
#         :param required:
#         :param nullable:
#         :param cast: Allow casting of value when deserializing.
#             Otherwise, not-exact datatypes causes ValidationError.
#         :param key:
#         :param attr:
#         :param default:
#         :param validators:
#         :param serializable:
#         :param deserializable:
#         :param description:
#         """
#         self.cls = cls
#         self.required = required
#         self.nullable = nullable
#         self.cast = cast
#         self.key = key
#         self.attr = attr
#         self.default = default
#         self.validators = validators or []
#         self.serializable = serializable
#         self.deserializable = deserializable
#         self.description = description
#
#     def validate(
#             self,
#             data: TSerialized,
#             context: Context,
#     ):
#         """
#         :param value:
#         :param context:
#         :return:
#         """
#         if data is None and not self.nullable:
#             raise ValidationError('Field is not nullable')
#
#         # if not isinstance(data)
#         # if self.cast:
#         #     try:
#
#         # cls = self.factory or context.default_factory
#         #
#         # try:
#         #     value = cls(value)
#         # except ValueError:
#         #     raise ValidationError('Invalid type')
#         # except:
#         #     j = 2
#         #     raise
#
#         for validator in self.validators:
#             validator(data)
#
#     @abstractmethod
#     def serialize(
#             self,
#             obj: TObject,
#             context: Context,
#     ) -> TSerialized:
#         """
#         :param obj:
#         :param context:
#         :return:
#         """
#         raise NotImplementedError
#
#     def serialize_many(
#             self,
#             obj_list: Iterable[TObject],
#             context: Context,
#     ) -> List[TSerialized]:
#         """
#         :param obj_list:
#         :param context:
#         :return:
#         """
#         values = []
#         # errors = ValidationError()
#
#         with CatchValidationErrors() as errors:
#             for i, obj in enumerate(obj_list):
#                 try:
#                     value = self.serialize(
#                         obj=obj,
#                         context=context,
#                     )
#                 except ValidationError as e:
#                     errors.add(i, e)
#                 else:
#                     values.append(value)
#
#         # if errors:
#         #     raise errors
#
#         return values
#
#     @abstractmethod
#     def deserialize(
#             self,
#             data: TSerialized,
#             context: Context,
#     ) -> TObject:
#         """
#         :param data:
#         :param context:
#         :return:
#         """
#         raise NotImplementedError
#
#     def deserialize_many(
#             self,
#             data_list: Iterable[TSerialized],
#             context: Context,
#     ) -> List[TObject]:
#         """
#         :param data_list:
#         :param context:
#         :return:
#         """
#         raise NotImplementedError
#
#     # def cast(self, context: Context, data: TData) -> TObject:
#     #     """
#     #     Cast
#     #     :param context:
#     #     :param data:
#     #     :return:
#     #     """
#     #     return self.cls(data)



from types import SimpleNamespace
from typing import Callable, Optional, Dict, Any, NamedTuple

from .validation import ValidationError


# Unless otherwise specified, fields will be deserialized into a
# SimpleNamespace, as its constructor takes **kwargs, which are available
# afterwards as attributes on the object (similar to how dataclasses works).
default_factory = SimpleNamespace
# default_factory = NamedTuple  # TODO


class Context(object):
    """
    The context implements and describes a common set of options
    used when serializing and deserializing.
    """

    def __init__(
            self,
            options: Optional[Dict[Any, Any]] = None,
            include_none: bool = True,
            strict: bool = True,
            ordered: bool = False,
            default_factory: Callable = default_factory,
            default_date_format: str = '',  # TODO
            default_time_format: str = '',  # TODO
            default_datetime_format: str = '',  # TODO
    ):
        """
        :param options: Arbitrary dict to store custom options.
            Accessible by Schema and Field when serializing and deserializing.
        :param include_none: Whether to include None values when serializing.
        :param ordered: Whether to keep the order of keys when serializing.
        :param strict: Default strict for Schemas, unless
        :param ordered:
        :param default_factory:
        :param default_date_format:
        :param default_time_format:
        :param default_datetime_format:
        """
        self.options = options or {}
        self.include_none = include_none
        self.ordered = ordered
        self.default_factory = default_factory
        self.default_date_format = default_date_format
        self.default_time_format = default_time_format
        self.default_datetime_format = default_datetime_format

    def create_validation_error(self) -> ValidationError:
        """
        Creates a new, empty instance of ValidationError.

        If replacing the default ValidationError class, make sure you return
        a type that is a subclass of ValidationError.
        """
        return ValidationError()


# -- Singletons --------------------------------------------------------------

default_context = Context()



from typing import Optional, Any

from ..field import Field
from ..context import Context


class String(Field[str, str]):

    def __init__(self, **kwargs):
        super().__init__(cls=str, **kwargs)

    def cast(self, value: Any, context: Context) -> str:
        if isinstance(value, bytes):
            return value.decode('utf8')
        return super().cast(value, context)



from inspect import getmembers
from functools import cached_property, lru_cache
from typing import (
    Iterable, Dict as Dict_, Optional, Any, Tuple,
    Callable, Generic, Protocol, Type, Union,
    NamedTuple)

from .fields import Dict
from .field import Field, TObject
from .context import Context, default_context


# -- Types -------------------------------------------------------------------


# Type of serialized schema
from .validation import collect_validation_errors, ValidationError

TSchema = Dict_[str, Any]

# TODO
TFactory = Callable[[Context, TSchema], Any]


class ObjectFactory(Protocol):
    def __call__(self, **kwargs: Any) -> TObject: ...


# -- Schema ------------------------------------------------------------------


# class Schema(Dict):  # TODO
class Schema(Dict, Generic[TObject]):
    """
    TODO
    """

    def __init__(
            self,
            factory: Optional[Union[ObjectFactory, Type[TObject]]] = None,
            only: Optional[Iterable[str]] = None,
            strict: bool = True,
            **kwargs,
    ):
        """
        :param factory:
        :param strict:
        :param only:
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.factory = factory
        self.strict = strict
        self.only = only or ()

    @cached_property
    def fields(self) -> 'FieldContainer':
        """
        :return:
        """
        return FieldContainer(self)

    def cast(self, value: Any, context: Context) -> Any:
        return value

    def serialize(
            self,
            obj: Any,
            context: Context = default_context,
    ) -> TSchema:
        """
        :param obj:
        :param context:
        :return:
        """

        # Schema as dict, where key = field.key, and value is object
        schema_dict = {}

        with collect_validation_errors() as errors:

            for field in self.fields.serializable:
                # actual_key = field.key if field.key is not None else name
                # actual_attr = field.attr if field.attr is not None else name
                try:
                    if hasattr(obj, field.attr):
                        field_value = getattr(obj, field.attr)
                    else:
                        field_value = field.default_value
                except TypeError as e:
                    j = 2
                    # if self.strict and field.required:
                    # raise ValidationError('')

                if field_value is None and not context.include_none:
                    continue

                try:
                    schema_dict[field.key] = field.serialize(
                        obj=field_value,
                        context=context,
                    )
                except field.ValidationError as e:
                    errors.add(field.key, e)

        return schema_dict

    def deserialize(
            self,
            data: Optional[Dict_[str, Any]],
            context: Context = default_context,
    ) -> TObject:
        """
        :param data:
        :param context:
        :return:
        """
        if data is None:
            return None

        # Deserialize and validate using super class
        dict_data = super().deserialize(data, context)

        # Kwargs for object constructor, where key is field 'attr' and
        # value is the deserialized and validated value.
        kwargs = {}

        with collect_validation_errors() as errors:

            for field in self.fields.deserializable:
                if field.key in dict_data:
                    value = dict_data[field.key]
                elif self.strict:
                    errors.add(field.key, 'This key is required')
                    continue
                else:
                    value = field.default_value

                try:
                    kwargs[field.attr] = field.deserialize(
                        data=value,
                        context=context,
                    )
                except field.ValidationError as e:
                    errors.add(field.key, e)

            if self.strict:
                # TODO
                pass

        # with collect_validation_errors() as errors:
        #     for field_name, field in self.fields.deserializable:
        #         actual_key = field.key if field.key is not None else field_name  # noqa: E501
        #         actual_attr = field.attr if field.attr is not None else field_name  # noqa: E501
        #
        #         if actual_key in data:
        #             value = data[actual_key]
        #         elif self.strict:
        #             errors.add(actual_key, 'This key is required')
        #             continue
        #         else:
        #             value = field.default_value
        #
        #         try:
        #             obj_kwargs[actual_attr] = field.deserialize(value, context)
        #         except field.ValidationError as e:
        #             errors.add(actual_key, e)
        #
        # # Raise ValidationError, if any
        # if errors:
        #     raise errors

        # Instantiate object, passing **kwargs
        if self.factory is not None:
            return self.factory(**kwargs)
        else:
            return context.default_factory(**kwargs)


# -- Transformations ---------------------------------------------------------  # TODO


class FieldProxy(NamedTuple):
    field: Field
    name: str

    def __getattr__(self, item):
        return getattr(self.field, item)

    @property
    @lru_cache
    def key(self) -> str:
        return self.field.key if self.field.key is not None else self.name

    @property
    @lru_cache
    def attr(self) -> str:
        return self.field.attr if self.field.attr is not None else self.name


class FieldContainer(object):
    """
    TODO
    """
    def __init__(self, schema: Schema):
        """
        :param schema:
        """
        self.schema = schema

    def __iter__(self) -> Iterable[FieldProxy]:
        """
        :return:
        """
        return (
            FieldProxy(field=obj, name=name)
            for name, obj in getmembers(self.schema)
            if isinstance(obj, Field)
            and self.should_include(name)
        )

    @cached_property
    def serializable(self) -> Tuple[FieldProxy, ...]:
        """
        :return:
        """
        return tuple(f for f in self if f.serializable)

    @cached_property
    def deserializable(self) -> Tuple[FieldProxy, ...]:
        """
        :return:
        """
        return tuple(f for f in self if f.deserializable)

    def should_include(self, field_name: str) -> bool:
        """
        :return:
        """
        return not self.schema.only or field_name in self.schema.only



from datetime import time

from ..context import Context
from ..field import Field
from .string import String


# class Time(String):
#
#     def __init__(
#             self,
#             timezone_aware: bool = None,
#             **kwargs,
#     ):
#         """
#         :param kwargs: **kwargs for Field.__init__()
#         """
#         super().__init__(cls=time, **kwargs)
#
#     def deserialize(
#             self,
#             data: str,
#             context: Context = default_context,
#     ) -> time:




from typing import Any

from ..field import Field
from ..context import Context


class Dict(Field[dict, dict]):

    def __init__(self, **kwargs):
        """
        :param kwargs: **kwargs for Field.__init__()
        """
        kwargs.setdefault('default', dict)
        super().__init__(cls=dict, **kwargs)

    def cast(self, value: Any, context: Context) -> Field[dict, dict]:
        if not isinstance(value, dict):
            raise self.ValidationError(f'Can not cast to {self.cls}: {value}')
        return self.cls(value)



from ..field import Field


class DateTime(Field[int, int]):

    def __init__(self, timezone_aware: bool = True, **kwargs):
        """
        :param kwargs: **kwargs for Field.__init__()
        """
        super().__init__(cls=int, **kwargs)
        # super().__init__(cls=int, default=int, **kwargs)



from typing import List as _List

from ..field import Field
from ..context import Context


class List(Field[_List, _List]):

    def __init__(self, **kwargs):
        super().__init__(cls=int, **kwargs)

    def serialize(
            self,
            obj: list,
            context: Context,
    ) -> list:
        """
        :param obj:
        :param context:
        :return:
        """
        return list(obj)

    def deserialize(
            self,
            data: list,
            context: Context,
    ) -> list:
        """
        :param data:
        :param context:
        :return:
        """
        return list(data)



from ..field import Field


class Boolean(Field[bool, bool]):

    def __init__(self, **kwargs):
        """
        :param kwargs: **kwargs for Field.__init__()
        """
        super().__init__(cls=bool, **kwargs)



# from abc import abstractmethod
# from typing import Any, Type, Iterable, Callable, Generic
#
# from .context import Context
# from .validation import ValidationError
#
#
# TAction = Callable[[Any, Context], Iterable[Any]]
#
# TSerialize = Callable[[Any, Context], Iterable[Any]]
#
#
# def b(a: int, b: Context):
#     return (i for i in range(1))
#
#
# def a() -> Iterable[int]:
#     return b(1, Context())
#
#
# # class Action(Generic[TObject]):
# #
# #     @abstractmethod
# #     def __call__(self, value: Any, context: Context) -> Any:
# #         """
# #         TODO
# #         """
# #         raise NotImplementedError
#
#
# # class Transformation(TTransformation):
# #     """
# #     TODO
# #     """
# #     def __call__(self, value: Any, context: Context) -> Any:
# #         """
# #         TODO
# #         """
# #         return self.apply(value, context)
# #
# #     @abstractmethod
# #     def apply(self, value: Any, context: Context) -> Any:
# #         """
# #         :param value:
# #         :param context:
# #         :returns:
# #         """
# #         raise NotImplementedError
#
#
# class RequireType(TAction):
#     def __init__(self, type: Type[Any], nullable: bool = False):
#         self.type = type
#         self.nullable = nullable
#
#     def __call__(self, value: Any, context: Context) -> Any:
#         """
#         TODO
#         """
#         if not isinstance(value, self.type):
#             raise ValidationError()
#
#
# class Cast(TAction):
#     def __init__(self, type: Type[Any]):
#         self.type = type
#
#     def __call__(self, value: Any, context: Context) -> Any:
#         """
#         TODO
#         """
#         return self.type(value)
#         # raise Exception('asd!')
#         # if not isinstance(value, self.type):
#         # return self.apply(value, context)
#
#     # def __init__(self, type: Type[Any]):
#     #     self.type = type
#     # def apply(self, value: Any, context: Context) -> Any:
#     #     return value
#
#
# class Default(TAction):
#     def __init__(self, default: Any):
#         self.default = default
#     def __call__(self, value: Any, context: Context) -> Any:
#         """
#         TODO
#         """
#         raise Exception('asd!')
#
#
# class Pipeline(TAction):
#     """
#     TODO
#     """
#     def __init__(self, *actions: TAction):
#         """
#         :param actions:
#         """
#         self.actions = actions
#
#     def __call__(self, value: Any, context: Context) -> Any:
#         """
#         Left-fold value through all transformations.
#
#         :param value:
#         :param context:
#         :returns: Transformed value
#         """
#         for action in self.actions:
#             value_transformed = action(value, context)
#             if value_transformed is not None:
#                 value = value_transformed
#
#         return value
#
#
# def pipe(*actions: TAction) -> Pipeline:
#     return Pipeline(*actions)
